#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import locale
import i18n
import logging

LOG = logging.getLogger('hawthorn')
default_language, _ = locale.getdefaultlocale()

i18n.set('fallback', 'en_US')

def init_i18n_locales_path():
    lan_path = os.path.join(os.path.dirname(__file__), 'locales')
    LOG.info("Initialize i18n language locales path %s", lan_path)
    if not os.path.exists(lan_path):
        # lan_path = os.path.join(os.path.dirname(__file__), 'locales')
        # TODO: process if this deployment environment occured
        pass
    if os.path.exists(lan_path):
        LOG.info("add language locales path: %s", lan_path)
        i18n.load_path.append(lan_path)

def init_i18n(language: str = None):
    LOG.info("Initialize i18n configuration")
    if not language:
        language = default_language
    i18n.set('locale', language)
    i18n.set('fallback', 'en_US')

init_i18n()
init_i18n_locales_path()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from .resultcodes import RESULT_CODE
from . import JsonEncoder

class GeneralResponseObject(object):
    """
    General Response Object
    """
    class GeneralResponseObjectError(TypeError):
        pass

    def __init__(self, code=-2, message='Unsupported operation'):
        self.code = code
        self.data = None
        self.message = message

    def is_success(self):
        return self.code == RESULT_CODE.OK

    def as_dict(self):
        result = {
            'code': self.code,
            'message': self.message
        }
        if self.data is not None:
            result['data'] = self.data
        return result

    def encode_json(self):
        obj = self.as_dict()
        return json.dumps(obj, cls=JsonEncoder)

class ListResponseObject(object):
    """
    List Data Response Object
    """
    class ListResponseObjectError(TypeError):
        pass

    def __init__(self, code=-2, message='Unsupported operation'):
        self.code = code
        self.total = 0
        self.page = 0
        self.pageSize = 0
        self.data = []
        self.message = message

    def is_success(self):
        return self.code == RESULT_CODE.OK

    def as_dict(self):
        result = {
            'code': self.code,
            'total': self.total,
            'page': self.page,
            'pageSize': self.pageSize,
            'data': self.data,
            'message': self.message
        }
        return result

    def encode_json(self):
        obj = self.as_dict()
        return json.dumps(obj, cls=JsonEncoder)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import tornado.web
import tornado.httputil
import sqlalchemy
import i18n
from http import HTTPStatus
from typing import Iterable
from collections.abc import Callable
from wtforms import Form
from hawthorn.asynchttphandler import GeneralTornadoHandler
from hawthorn.modelutils import ModelBase, model_columns, get_model_class_name
from hawthorn.dbproxy import DbProxy
from ..valueobjects.resultcodes import RESULT_CODE
from ..valueobjects.responseobject import GeneralResponseObject

from .restdescriptor import RESTfulAPIWraper
from .utils import read_request_parameters, get_locale_params

LOG = logging.getLogger('services.generalmodelapi.apihandlers')

class ModelPageDescriptorsApiHandler():
    """
    Model page descriptor API handler wrapper
    """
    def __init__(self, get_page_descriptor_model: Callable[[str], RESTfulAPIWraper]):
        self.get_page_descriptor_model = get_page_descriptor_model

    async def handler_get(self, handler: GeneralTornadoHandler, request: tornado.httputil.HTTPServerRequest):
        """
        API handler of get single model data
        """
        filters = read_request_parameters(request)
        locale_params = get_locale_params(request)
        
        result = GeneralResponseObject(RESULT_CODE.DATA_DOES_NOT_EXISTS, message=i18n.t('basic.data_not_exists', **locale_params))
        while result.code != RESULT_CODE.OK:
            page_route = str(filters.get('pathname', ''))
            route_pieces = page_route.split('/')
            if not page_route or not route_pieces:
                result.code = RESULT_CODE.INVALID_PARAM
                result.message = i18n.t('basic.route_path_name_should_not_empty', **locale_params)
                break

            page_name = route_pieces[len(route_pieces)-1]
            w = self.get_page_descriptor_model(page_name)
            if not w:
                result.code = RESULT_CODE.INVALID_PARAM
                result.message = i18n.t('basic.could_not_recognize_current_page', **locale_params)
                break
            if not w.cls:
                result.code = RESULT_CODE.INTERNAL_ERROR
                result.message = f'No model specified for page name: {page_name}' # i18n.t('basic.could_not_recognize_current_page', **locale_params)
                break

            result.code = RESULT_CODE.OK
            result.message = i18n.t('basic.success', **locale_params)
            result.data = w.destriptor()
            break
        
        return result.encode_json()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import i18n
from wtforms import Form
from hawthorn.modelutils import ModelBase, model_columns
from ..valueobjects.resultcodes import RESULT_CODE
from ..valueobjects.responseobject import GeneralResponseObject

def prepare_item_fields(form_item: Form):
    # if not fields or not isinstance(fields, dict):
    #     if isinstance(fields, list):
    #         return { v: v for v in fields }
    #     rfields = {}
    #     default_cols, pk = model_columns(self.model)
    #     for col in default_cols:
    #         if col != pk and col not in DEFAULT_SKIP_FIELDS:
    #             rfields[col] = col
    #     return rfields
    # return fields
    rfields = {}
    for field in form_item:
        if field.name not in rfields and not field.flags.hidden:
            rfields[field.name] = field.id if field.id else field.name

    return rfields

def save_form_fields(form_item: Form, model: ModelBase, ignore_empty=True):
    columns, pk = model_columns(model)
    for field in form_item:
        field_name = field.id if field.id else field.name
        if field_name == pk:
            continue
        if hasattr(model, field_name):
            if not field.data:
                if ignore_empty and isinstance(field.data, str) or isinstance(field.data, bytes):
                    continue
            # if model has set_<property> method, call set_<property> to set model field data
            if hasattr(model, 'set_' + field_name) and callable(getattr(model, 'set_' + field_name, None)):
                getattr(model, 'set_' + field_name)(field.data)
            else:
                setattr(model, field_name, field.data)

def validate_form(form_item: Form, **kwargs):
    """
    Validate form data
    @param form_item: Form input data
    @param locale=en_US [optional]
    """
    locale_params = {}
    language = kwargs.pop('locale', None)
    if language:
        locale_params['locale'] = language
    if not form_item.validate():
        errs = form_item.errors
        if errs:
            errslices = []
            for errelements in errs.values():
                errslices.append(';'.join(errelements))
            errtext = ';'.join(errslices)
        else:
            errtext = i18n.t('basic.invalid_param', **locale_params)
        return GeneralResponseObject(RESULT_CODE.INVALID_PARAM, errtext)
    return GeneralResponseObject(RESULT_CODE.OK, message=i18n.t('basic.success', **locale_params))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json

class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytes):
            return obj.decode('utf-8')
        return json.JSONEncoder.default(self, obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from wtforms import Form
from hawthorn.modelutils import ModelBase, get_model_class_name
from .accesscontrol.authorizedsession import SESSION_UID_KEY
from .restdescriptor import RESTfulAPIWraper
from .register import register_model_general_api_handlers, register_model_page_descriptor_api_handler

__all__ = ['restful_api', 'SESSION_UID_KEY']

LOG = logging.getLogger('hawthorn.restfulwrapper')

__restful_api_wrappers: dict[str, RESTfulAPIWraper] = {}
__page_descriptor_model_wrappers: dict[str, RESTfulAPIWraper] = {}
__restful_api_registered = False

def restful_api(endpoint: str = '', title='', form: Form = None, **kwargs):
    """
    wrapper model as RESTful API endpoint

    usage:  @restful_api(endpoint)
            class ModelName():
                def __init__(self):
                    pass
    """
    def decorator(cls: ModelBase):
        uri = endpoint
        if not uri:
            uri = '/api/' + get_model_class_name(cls).lower() + 's'
            LOG.info('treat %s as RESTful URI:%s', get_model_class_name(cls), uri)
        if uri and uri not in __restful_api_wrappers:
            w = RESTfulAPIWraper(uri, cls, title=title, form=form)
            __restful_api_wrappers[uri] = w
            route_pieces = uri.split('/')
            page_name = route_pieces[len(route_pieces)-1]
            __page_descriptor_model_wrappers[page_name] = w
        else:
            raise ValueError("Duplicate registering model RESTful URI '%s'" % (uri))
        
        return cls
    return decorator

def get_all_restful_api_wrappers():
    return __restful_api_wrappers

def get_page_descriptor_model(pathname: str) -> RESTfulAPIWraper:
    if pathname in __page_descriptor_model_wrappers:
        return __page_descriptor_model_wrappers[pathname]
    return None

def register_restful_apis():
    global __restful_api_registered
    if not __restful_api_wrappers or __restful_api_registered:
        return
    for endpoint, w in __restful_api_wrappers.items():
        if endpoint and w.cls:
            # LOG.info('registing RESTful endpoint %s', endpoint)
            register_model_general_api_handlers(w.cls, endpoint, form=w.form)
    register_model_page_descriptor_api_handler(get_page_descriptor_model)
    __restful_api_registered = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sqlalchemy import Column, Integer, String

from hawthorn.modelutils import ModelBase
from hawthorn.modelutils.behaviors import ModifyingBehevior

class _MigrationProgressTable(ModelBase):
    __tablename__ = 'sys_migration_progress'

    id = Column('id', Integer, primary_key=True, autoincrement=True)
    app = Column('app', String(64), index=True)
    version = Column('version', String(64), index=True)
    module = Column('module', String(64), index=True)
    value = Column('value', String(256))
    progress = Column('progress', Integer)
    total = Column('total', Integer)
    cost_seconds = Column('cost_seconds', Integer)
    left_seconds = Column('left_seconds', Integer)

class MigrationProgress(_MigrationProgressTable, ModifyingBehevior):
    """
    Migration module progress
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import tornado.web
import tornado.httputil
import tornado.gen
import asyncio
import sqlalchemy
import i18n
import traceback
from http import HTTPStatus
from typing import Iterable
from wtforms import Form
from hawthorn.asynchttphandler import GeneralTornadoHandler, request_body_as_json
from hawthorn.session import TornadoSession
from hawthorn.exceptionreporter import ExceptionReporter
from hawthorn.modelutils import ModelBase, model_columns, get_model_class_name
from hawthorn.dbproxy import DbProxy
from . import SESSION_UID_KEY
from ..formvalidation.formutils import validate_form, save_form_fields
from ..valueobjects.resultcodes import RESULT_CODE
from ..valueobjects.responseobject import GeneralResponseObject, ListResponseObject

from .utils import format_model_query_conditions, dump_model_data, read_request_parameters, get_locale_params, get_listquery_pager_info, get_listquery_sort_info

LOG = logging.getLogger('services.generalmodelapi.apihandlers')

class ModelRESTfulHandler(tornado.web.RequestHandler):
    """
    Model API handler wrapper
    """
    def initialize(self, model: ModelBase, form: Form = None, ac=[]):
        self.model = model
        self.form = form
        self._ac = []
        self.session = TornadoSession(self)
        self.set_access_control(ac)

    def set_access_control(self, ac):
        self._ac = []
        if ac:
            if isinstance(ac, list):
                for f in ac:
                    if callable(f):
                        self._ac.append(['', f])
            elif isinstance(ac, dict):
                for k, f in ac:
                    if callable(f):
                        self._ac.append([k, f])
            elif callable(ac):
                self._ac.append(['', ac])
                
    async def check_access_control(self):
        is_reject = False
        reject_reason = ''
        for o in self._ac:
            f = o[1]
            try:
                if tornado.gen.is_coroutine_function(f) or asyncio.iscoroutinefunction(f):
                    ac_result, msg = await f(self)
                else:
                    ac_result, msg = f(self)
            except Exception as e:
                LOG.warning('verify %s access control %s excepted with error:%s traceback:%s', self.request.path, str(o[0]), str(e), traceback.format_exc())
                ac_result = False
                msg = str(e)
                ExceptionReporter().report(key='ACL-' + self.request.path, typ='HTTP', 
                    endpoint=self.request.path,
                    method=self.request.method,
                    inputs=str(o[0]),
                    outputs='',
                    content=str(e),
                    level='ERROR'
                )
            if not ac_result:
                reject_reason = str(msg)
                LOG.warning('verify %s access control %s failed with error:%s', self.request.path, str(o[0]), reject_reason)
                is_reject = True
                break
        return is_reject, reject_reason
    
    async def prepare(self):
        is_reject, reject_reason = await self.check_access_control()
        if is_reject:
            self.write(reject_reason)
            self.set_status(HTTPStatus.FORBIDDEN)
            self.finish()
            return

    async def get(self, id: str|int = None):
        """
        API handler of get single model data
        """
        filters = read_request_parameters(self.request)
        locale_params = get_locale_params(self.request)
        if id:
            columns, pk = model_columns(self.model)
            if id not in ['fetchOne', 'findOne']:
                filters[pk] = id
            result = GeneralResponseObject(RESULT_CODE.DATA_DOES_NOT_EXISTS, message=i18n.t('basic.data_not_exists', **locale_params))
            while result.code != RESULT_CODE.OK:
                filter_conds, err_msg = format_model_query_conditions(self.model, filters=filters)
                if err_msg:
                    result.code = RESULT_CODE.INVALID_PARAM
                    result.message = err_msg
                    break

                m = await DbProxy().find_item(self.model, filters=filter_conds)
                if m:
                    result.code = RESULT_CODE.OK
                    result.message = i18n.t('basic.success', **locale_params)
                    if hasattr(m, 'as_dict'):
                        result.data = getattr(m, 'as_dict')()
                    else:
                        result.data = dump_model_data(m, columns=columns)
                break
        else:
            result = await self.do_get_list(filters, locale_params)
        
        self.set_header('Content-Type', 'application/json')
        self.write(result.encode_json())
        self.finish()

    async def do_get_list(self, params: dict, locale_params: dict):
        """
        API handler of get model list data by filter condition and pagination
        """
        filters = params.get('filter', {})
        fields = params.get('fields', [])
        if not filters or not isinstance(filters, dict):
            filters = {}
        limit, offset = get_listquery_pager_info(params)
        orderby, direction = get_listquery_sort_info(params)

        result = ListResponseObject(RESULT_CODE.DATA_DOES_NOT_EXISTS, message=i18n.t('basic.data_not_exists', **locale_params))
        while result.code != RESULT_CODE.OK:
            filter_conds, err_msg = format_model_query_conditions(self.model, filters=filters)
            if err_msg:
                result.code = RESULT_CODE.INVALID_PARAM
                result.message = err_msg
                break

            rows, total = await DbProxy().query_list(self.model, filters=filter_conds, limit=limit, offset=offset, sort=orderby, direction=direction, selections=fields)
            
            if not fields:
                if hasattr(self.model, 'as_dict'):
                    formatted_rows = []
                    for row in rows:
                        m = self.model()
                        for k, v in row.items():
                            setattr(m, k, v)
                        formatted_rows.append(getattr(m, 'as_dict')())
                    rows = formatted_rows

            result.code = RESULT_CODE.OK
            result.message = i18n.t('basic.success', **locale_params)
            result.total = total
            result.data = rows
            if limit:
                result.pageSize = limit
                result.page = int(offset/limit) + 1
        return result

    async def post(self, id: str|int = None):
        """
        API handler of add model instance data
        """
        if id:
            self.set_status(HTTPStatus.NOT_FOUND)
            self.finish()
            return
        locale_params = get_locale_params(self.request)
        if not self.form:
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR, '%s form not configured for general form handler, please specify form when register model general api handlers' % (get_model_class_name(self.model)))
            self.finish()
            return
        inputs = request_body_as_json(self.request)
        if not inputs:
            self.set_status(HTTPStatus.NO_CONTENT, i18n.t('basic.invalid_param', **locale_params))
            self.finish()
            return
        form_items = []
        if isinstance(inputs, list):
            for ele in inputs:
                form_items.append(self.form(formdata=None, data=ele, meta={'csrf': False}))
        else:
            form_items.append(self.form(formdata=None, data=inputs, meta={'csrf': False}))
            
        insert_models = []
        for form_item in form_items:
            result = validate_form(form_item)
            if not result.is_success():
                self.set_header('Content-Type', 'application/json')
                self.write(result.encode_json())
                self.finish()
                return

            m = self.model()
            if hasattr(m, 'set_session_uid'):
                session_uid = self.session[SESSION_UID_KEY]
                setattr(m, 'set_session_uid', session_uid)
            save_form_fields(form_item, m, ignore_empty=True)
            insert_models.append(m)

        result = GeneralResponseObject(code=RESULT_CODE.FAIL, message=i18n.t('basic.create_data_failed', **locale_params))
        
        # m = self.model()
        # if hasattr(m, 'set_session_uid'):
        #     session_uid = self.session[SESSION_UID_KEY]
        #     setattr(m, 'set_session_uid', session_uid)
        # save_form_fields(form_item, m, ignore_empty=True)

        try:
            res = await DbProxy().insert_items(insert_models, auto_flush=True)
            if res:
                result.code = RESULT_CODE.OK
                result.message = i18n.t('basic.success', **locale_params)
                resp_data = []
                for m in insert_models:
                    if hasattr(m, 'as_dict'):
                        resp_data.append(getattr(m, 'as_dict')())
                    else:
                        resp_data.append(dump_model_data(m))
                if isinstance(inputs, list):
                    result.data = resp_data
                else:
                    result.data = resp_data[0]
                LOG.info('insert [%s] succeed', get_model_class_name(self.model))
            else:
                LOG.warning('insert [%s] failed', get_model_class_name(self.model))
        except Exception as e:
            LOG.error('insert [%s] failed with error:%s', get_model_class_name(self.model), str(e))
            result.code = RESULT_CODE.FAIL
            result.message = str(e)
        
        self.set_header('Content-Type', 'application/json')
        self.write(result.encode_json())
        self.finish()
    
    async def put(self, id: str|int = None):
        if not id:
            self.set_status(HTTPStatus.NOT_FOUND)
            self.finish()
            return
        await self.do_edit(id, True)
    
    async def patch(self, id: str|int = None):
        if not id:
            self.set_status(HTTPStatus.NOT_FOUND)
            self.finish()
            return
        await self.do_edit(id, False)

    async def do_edit(self, id: str|int, ignore_empty: bool):
        """
        API handler of edit model instance data
        """
        locale_params = get_locale_params(self.request)
        if not self.form:
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR, '%s form not configured for general form handler, please specify form when register model general api handlers' % (get_model_class_name(self.model)))
            self.finish()
            return
        inputs = request_body_as_json(self.request)
        columns, pk = model_columns(self.model)
        inputs[pk] = id
        form_item = self.form(formdata=None, data=inputs, meta={ 'csrf': False })
        result = validate_form(form_item)
        while result.is_success():
            form_pk_id = id
            m = await DbProxy().find_item(self.model, {getattr(self.model, pk)==form_pk_id})
            if not m:
                LOG.warning('get %s [%s] info failed while data does not extsts', get_model_class_name(self.model), str(form_pk_id))
                result = GeneralResponseObject(code=RESULT_CODE.DATA_DOES_NOT_EXISTS, message=i18n.t('basic.data_not_exists', **locale_params)).encode_json()
                break
            if hasattr(m, 'set_session_uid'):
                session_uid = self.session[SESSION_UID_KEY]
                setattr(m, 'set_session_uid', session_uid)
            save_form_fields(form_item, m, ignore_empty=ignore_empty)

            result = GeneralResponseObject(code=RESULT_CODE.FAIL, message=i18n.t('basic.edit_data_failed', **locale_params))
            try:
                res = await DbProxy().update_item(m)
                if res:
                    result.code = RESULT_CODE.OK
                    result.message = i18n.t('basic.success', **locale_params)
                    if hasattr(m, 'as_dict'):
                        result.data = getattr(m, 'as_dict')()
                    else:
                        result.data = dump_model_data(m, columns=columns)
                    LOG.info('edit %s [%s] succeed', get_model_class_name(self.model), str(form_pk_id))
                else:
                    LOG.warning('edit %s [%s] failed', get_model_class_name(self.model), str(form_pk_id))
            except Exception as e:
                LOG.error('edit %s [%s] failed with error:%s', get_model_class_name(self.model), str(form_pk_id), str(e))
                result.code = RESULT_CODE.FAIL
                result.message = str(e)
            break
        self.set_header('Content-Type', 'application/json')
        self.write(result.encode_json())
        self.finish()

    async def delete(self, id: str|int = None):
        """
        API handler of delete model instance data
        """
        if not id:
            self.set_status(HTTPStatus.NOT_FOUND)
            self.finish()
            return
        # inputs = request_body_as_json(self.request)
        locale_params = get_locale_params(self.request)
        columns, pk = model_columns(self.model)
        form_pk_id = id
        if not form_pk_id:
            LOG.warning('delete %s while not giving any id to delete', get_model_class_name(self.model))
            return GeneralResponseObject(code=RESULT_CODE.INVALID_PARAM, message=i18n.t('basic.invalid_param', **locale_params)).encode_json()
        m = await DbProxy().find_item(self.model, {getattr(self.model, pk)==form_pk_id})
        if not m:
            LOG.warning('delete %s [%s] info failed while data does not extsts', get_model_class_name(self.model), str(form_pk_id))
            return GeneralResponseObject(code=RESULT_CODE.DATA_DOES_NOT_EXISTS, message=i18n.t('basic.data_not_exists', **locale_params)).encode_json()
        
        result = GeneralResponseObject(code=RESULT_CODE.FAIL, message=i18n.t('basic.delete_data_failed', **locale_params))
        try:
            res = False
            if False and hasattr(m, 'obsoleted'):
                if hasattr(m, 'set_session_uid'):
                    session_uid = self.session[SESSION_UID_KEY]
                    setattr(m, 'set_session_uid', session_uid)
                setattr(m, 'obsoleted', 1)
                res = await DbProxy().update_item(m)
            else:
                res = await DbProxy().del_item(m)
            if res:
                result.code = RESULT_CODE.OK
                result.message = i18n.t('basic.success', **locale_params)
                LOG.info('delete %s [%s] succeed', get_model_class_name(self.model), str(form_pk_id))
            else:
                LOG.warning('delete %s [%s] failed', get_model_class_name(self.model), str(form_pk_id))
        except Exception as e:
            LOG.error('delete %s [%s] failed with error:%s', get_model_class_name(self.model), str(form_pk_id), str(e))
            result.code = RESULT_CODE.FAIL
            result.message = str(e)
        return result.encode_json()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from wtforms.validators import StopValidation, ValidationError, Regexp
from hawthorn.dbproxy import DbProxy
from hawthorn.modelutils import ModelBase, MongoBase
from hawthorn.utilities import toint, tofloat

class DataExistsValidator(object):
    """
    验证数据是否存在

    :param model:
        The model to be checked.
    :param index_field:
        The field to be checked.
    :param name_field:
        The field to be displayed.
    :param message:
        Error message to raise in case of a validation error.
    """

    def __init__(self, model, index_field='id', name_field='name', 
                 message=None, 
                 should_exists=True, 
                 allows_empty=False,
                 ignore_obsoleted=False,
                 reference_field='id',
                 primary_field='id',
                 union_field=None,
                 fixed_field=None):
        self.model = model
        self.index_field = index_field
        self.name_field = name_field
        self.reference_field = reference_field
        self.primary_field = primary_field
        self.union_field = union_field
        self.fixed_field = fixed_field or {}
        self.message = message
        self.should_exists = should_exists
        self.allows_empty = allows_empty
        self.ignore_obsoleted = ignore_obsoleted

    async def __call__(self, form, field):
        if not field.data or isinstance(field.data, str) and not field.data.strip():
            if self.allows_empty:
                return
            if self.message is None:
                message = field.gettext('This field is required.')
            else:
                message = self.message

            field.errors[:] = []
            raise StopValidation(message)
        else:
            model = self.get_validate_model(form)
            index_field = self.get_validate_index_field(form)
            primary = (getattr(form, self.primary_field, None) and getattr(form, self.primary_field).data) or ''
            union = [{'field': field, 'data': (getattr(form, field, None) and getattr(form, field).data or '')} for field in self.union_field] if self.union_field else None
            if not model or not index_field:
                message = 'Field:%s value:%s not valid.' % (self.reference_field, str(getattr(form, self.reference_field).data))
                field.errors[:] = []
                raise StopValidation(message)
            if isinstance(model, MongoBase.my_metaclass):
                mongo_condition = {**self.fixed_field, index_field: field.data}
                if not self.ignore_obsoleted:
                    if hasattr(model, 'obsoleted'):
                        mongo_condition['obsoleted'] = False
                if primary:
                    mongo_condition[self.primary_field + '__ne'] = primary
                if union:
                    for u in union:
                        if u.get('field') and u.get('data'):
                            mongo_condition[u.get('field')] = u.get('data')
                try:
                    val = await DbProxy().query_all_mongo(model, mongo_condition)
                except Exception as e:
                    field.errors[:] = []
                    raise StopValidation(str(e))
            elif issubclass(model, ModelBase):
                pg_condition = [getattr(model, index_field) == field.data]
                if self.fixed_field:
                    for key, value in self.fixed_field.items():
                        if hasattr(model, key):
                            pg_condition.append(getattr(model, key) == value)
                if not self.ignore_obsoleted:
                    if hasattr(model, 'obsoleted'):
                        pg_condition.append(getattr(model, 'obsoleted') == 0)
                if primary:
                    pg_condition.append(getattr(model, self.primary_field) != primary)
                if union:
                    for u in union:
                        if u.get('field') and u.get('data'):
                            pg_condition.append(getattr(model, u.get('field')) == u.get('data'))
                try:
                    val = await DbProxy().find_item(model, pg_condition)
                except Exception as e:
                    field.errors[:] = []
                    raise StopValidation(str(e))

            message = None
            if self.should_exists:
                if not val:
                    if self.message:
                        message = self.message
                    else:
                        message = field.gettext('The field data does not exists.')
            else:
                if val:
                    if self.message:
                        message = self.message
                    else:
                        message = field.gettext('The data should not exists.')

            if message:
                field.errors[:] = []
                raise StopValidation(message)
    
    def get_validate_model(self, form):
        model = self.model
        if isinstance(model, dict):
            if getattr(form, self.reference_field).data in model:
                model = model[getattr(form, self.reference_field).data]
            else:
                return None
        return model
    
    def get_validate_index_field(self, form):
        index_field = self.index_field
        if isinstance(index_field, dict):
            if getattr(form, self.reference_field).data in index_field:
                index_field = index_field[getattr(form, self.reference_field).data]
            else:
                return None
        return index_field
    
    def get_validate_name_field(self, form):
        name_field = self.name_field
        if isinstance(name_field, dict):
            if getattr(form, self.reference_field).data in name_field:
                name_field = name_field[getattr(form, self.reference_field).data]
            else:
                return None
        return name_field
    
class DataDictsValidator(object):
    """
    验证数据字典是否合法，并可根据字典值为名称为数据查询拼装名称

    :param dicts:
        The dict to be checked.
    :param message:
        Error message to raise in case of a validation error.
    """

    def __init__(self, dicts, 
                 message=None,
                 multi_dicts=False,
                 values_formatter=None, 
                 should_exists=True, 
                 allows_empty=False,
                 reference_field='id'):
        self.values = dicts
        self.multi_dicts = multi_dicts
        self.message = message
        if values_formatter is None:
            values_formatter = self.default_values_formatter
        self.values_formatter = values_formatter
        self.should_exists = should_exists
        self.allows_empty = allows_empty
        self.reference_field = reference_field

    def __call__(self, form, field):
        if not field.data or isinstance(field.data, str) and not field.data.strip():
            if self.allows_empty:
                return
            if self.message is None:
                message = field.gettext('This field is required.')
            else:
                message = self.message

            field.errors[:] = []
            raise StopValidation(message)
        else:
            values = self.get_validate_model(form)
            if not values:
                message = 'Field:%s value:%s not valid.' % (self.reference_field, str(getattr(form, self.reference_field).data))
                field.errors[:] = []
                raise StopValidation(message)
                
            if field.data not in self.values:
                message = self.message
                if message is None:
                    message = field.gettext('Invalid value, must be one of: %(values)s.')

                raise ValidationError(message % dict(values=self.values_formatter(self.values)))
    
    @staticmethod
    def default_values_formatter(values):
        return ', '.join(str(x) for x in values.keys())

    def get_validate_model(self, form):
        values = self.values
        if self.multi_dicts:
            if getattr(form, self.reference_field).data in values:
                values = values[getattr(form, self.reference_field).data]
            else:
                return None
        return values

class PositiveRealValidators(object):
    """
    验证IntegerFiled 是否为正实数
    """
    def __init__(self, message=None):
        self.message = message

    def __call__(self, form, field):
        try:
            d = toint(field.data)
        except ValueError as e:
            d = tofloat(field.data)
        if d < 0:
            if self.message is None:
                message = field.gettext('This field is required positive integer.')
            else:
                message = self.message

            field.errors[:] = []
            raise StopValidation(message)

class DateTimeValidator(Regexp):
    """
    Validates the field against a user provided regexp.

    :param regex:
        The regular expression string to use. Can also be a compiled regular
        expression pattern.
    :param flags:
        The regexp flags to use, for example re.IGNORECASE. Ignored if
        `regex` is not a string.
    :param message:
        Error message to raise in case of a validation error.
    """

    def __init__(self, regex=r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$', flags=0, message=None):
        super().__init__(regex, flags, message)

    def __call__(self, form, field, message=None):
        match = self.regex.match(field.data or "")
        if match:
            return match

        if message is None:
            if self.message is None:
                message = field.gettext("Invalid datetime format.")
            else:
                message = self.message

        raise ValidationError(message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from hawthorn.asynchttphandler import GeneralTornadoHandler

SESSION_UID_KEY = 'uid'

def authorized_session_access_control(handler: GeneralTornadoHandler):
    ac_result = False
    ac_message = 'Reject'
    authed_uid = None
    if handler.session:
        authed_uid = handler.session[SESSION_UID_KEY]
    if authed_uid:
        ac_result = True
        ac_message = 'Authorized'
    
    return ac_result, ac_message



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import bcrypt

def get_current_timestamp():
    """
    Get current timestamp as milli seconds
    """
    return int(time.time() * 1000)

def generate_password(pwd: str):
    if isinstance(pwd, str):
        pwd = pwd.encode('utf-8')
    salt = bcrypt.gensalt()
    hashed = bcrypt.hashpw(pwd, salt)
    return hashed

def verify_password(pwd: str, password_hash: str):
    if isinstance(pwd, str):
        pwd = pwd.encode('utf-8')
    if isinstance(password_hash, str):
        password_hash = password_hash.encode('utf-8')
    if bcrypt.checkpw(pwd, password_hash):
        return True
    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sqlalchemy import Column
from wtforms import Form, Field
from wtforms.validators import DataRequired, NumberRange, Length, Regexp, AnyOf, Email, URL
from hawthorn.modelutils import ModelBase, model_columns, get_model_class_name
from lycium_rest.formvalidation.validators import DateTimeValidator

class RESTfulAPIWraper:
    def __init__(self, endpoint: str, cls: ModelBase, title: str = '', form: Form = None):
        self.endpoint = endpoint
        self.title = title
        self.cls = cls
        self.form = form
        self._descriptor = {}

    def destriptor(self, host: str=''):
        if not self._descriptor:
            self._descriptor = {
                'title': self.title if self.title else '',
                'cardBordered': True,
                'fetchDataURL': host + self.endpoint,
                'saveURL': host + self.endpoint + '/:id',
                'saveURLMethod': 'PATCH',
                'newURL': host + self.endpoint,
                'newURLMethod': 'POST',
                'viewURL': host + self.endpoint + '/:id',
                'viewURLMethod': 'GET',
                'deleteURL': host + self.endpoint + '/:id',
                'deleteURLMethod': 'DELETE',
                'editable': True,
                'rowKey': 'id',
                'pagination': {
                    'pageSize': 10,
                },
                'columns': []
            }
            if self.cls:
                columns, pk = model_columns(self.cls)
                formFieldsMapper = {}
                if self.form:
                    form = self.form()
                    for name, field in form._fields.items():
                        # field: Field = getattr(self.form, name)
                        colname = name
                        if field.id:
                            colname = field.id
                        formFieldsMapper[colname] = field
                            
                self._descriptor['rowKey'] = pk
                self._descriptor['columns'] = [self.generate_column_descriptor(colname, formFieldsMapper) for colname in columns]
        return self._descriptor

    def generate_column_descriptor(self, colname: str, formFieldsMapper: dict[str, Field]):
        column = {
            'key': colname,
            'name': colname,
            'valueType': 'text',
            'formItemProps': {'rules': []},
        }
        colfield: Column = getattr(self.cls, colname)
        if colfield.comment:
            column['description'] = colfield.comment
        if colfield.autoincrement:
            column['hideInForm'] = True
            column['readonly'] = True
            column['hideInSearch'] = True
        elif colfield.index:
            column['hideInSearch'] = False
        if colname in formFieldsMapper:
            field = formFieldsMapper[colname]
            if field.label:
                column['label'] = field.label.text
            if field.description:
                column['description'] = field.description
            for validator in field.validators:
                if isinstance(validator, DataRequired):
                    column['formItemProps']['rules'].append({'required': True, 'message': validator.message})
                elif isinstance(validator, DateTimeValidator):
                    column['valueType'] = 'dateTime'
                    column['sortable'] = True
                elif isinstance(validator, Regexp):
                    column['formItemProps']['rules'].append({'pattern': validator.regex.pattern, 'message': validator.message})
                elif isinstance(validator, Length) or isinstance(validator, NumberRange):
                    if validator.max > 0:
                        column['formItemProps']['rules'].append({'max': validator.max, 'message': validator.message})
                    if validator.min > 0:
                        column['formItemProps']['rules'].append({'min': validator.min, 'message': validator.message})
                    # if isinstance(validator, NumberRange):
                    #     column['valueType'] = 'number'
                elif isinstance(validator, AnyOf):
                    enumValues = validator.values
                    if isinstance(validator.values, dict):
                        column['valueType'] = 'select'
                        column['valueEnum'] = [{'value': k, 'text': v} for k, v in validator.values.items()]
                        enumValues = [k for k, _ in validator.values.items()]
                    column['formItemProps']['rules'].append({'enum': enumValues, 'message': validator.message})
                elif isinstance(validator, Email):
                    column['formItemProps']['rules'].append({'type': 'email', 'message': validator.message})
                elif isinstance(validator, URL):
                    column['formItemProps']['rules'].append({'type': 'url', 'message': validator.message})
        return column



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
import tornado.web
import tornado.httputil
import sqlalchemy
from typing import Iterable
from hawthorn.asynchttphandler import args_as_dict, request_body_as_json
from hawthorn.modelutils import ModelBase, model_columns
from hawthorn.utilities import toint

LOG = logging.getLogger('services.generalmodelapi.utils')

def format_model_query_conditions(model: ModelBase, filters: dict = {}, skip_non_existed_column=True):
    filter_conds = []
    err_messages = []
    if filters:
        for k, v in filters.items():
            if not hasattr(model, k):
                if not skip_non_existed_column:
                    err_messages.append(f'column {k} were not existed')
                continue
            model_field: sqlalchemy.Column = getattr(model, k)
            if isinstance(v, dict):
                for op, cv in v.items():
                    if '<>' == op or '!=' == op or '$ne' == op:
                        filter_conds.append(model_field!=cv)
                    elif '=' == op or '$eq' == op:
                        filter_conds.append(model_field==cv)
                    elif '>' == op or '$gt' == op:
                        filter_conds.append(model_field>cv)
                    elif '>=' == op or '$ge' == op:
                        filter_conds.append(model_field>=cv)
                    elif '<' == op or '$lt' == op:
                        filter_conds.append(model_field<cv)
                    elif '<=' == op or '$le' == op:
                        filter_conds.append(model_field<=cv)
                    elif 'contains' == op:
                        filter_conds.append(model_field.contains(str(cv)))
                    elif 'does not contain' == op:
                        filter_conds.append(model_field.notilike('%' + str(cv) + '%'))
                    elif 'begin with' == op:
                        filter_conds.append(model_field.startswith(str(cv)))
                    elif 'does not begin with' == op:
                        filter_conds.append(model_field.notilike(str(cv) + '%'))
                    elif 'end with' == op:
                        filter_conds.append(model_field.endswith(str(cv)))
                    elif 'does not end with' == op:
                        filter_conds.append(model_field.notilike('%' + str(cv)))
                    elif 'is null' == op or 'isnull' == op:
                        filter_conds.append(model_field.is_(None))
                    elif 'is not null' == op or 'isnotnull' == op:
                        filter_conds.append(model_field.is_not(None))
                    elif 'is empty' == op or 'isempty' == op:
                        filter_conds.append(model_field.is_(None))
                    elif 'is not empty' == op or 'isnotempty' == op:
                        filter_conds.append(model_field.is_not(None))
                    elif 'is between' == op:
                        if isinstance(cv, Iterable) and len(cv > 1):
                            filter_conds.append(model_field.between(cv[0], cv[1]))
                        else:
                            err_messages.append(f'Operator {op} value for column {k} were not iterable or compare values count less than 2')
                    elif 'is not between' == op:
                        if isinstance(cv, Iterable) and len(cv > 1):
                            filter_conds.append(~model_field.between(cv[0], cv[1]))
                        else:
                            err_messages.append(f'Operator {op} value for column {k} were not iterable or compare values count less than 2')
                    elif 'in' == op or 'is in' == 'op' or 'is in list' == 'op':
                        if isinstance(cv, Iterable):
                            filter_conds.append(model_field.in_(cv))
                        else:
                            err_messages.append(f'Operator {op} value for column {k} were not iterable')
                    elif 'not in' == op or 'is not in' == 'op' or 'is not in list' == 'op':
                        if isinstance(cv, Iterable):
                            filter_conds.append(model_field.not_in(cv))
                        else:
                            err_messages.append(f'Operator {op} value for column {k} were not iterable')
                    else:
                        err_messages.append(f'Operator {op} for column {k} were not supported')
                continue
            elif isinstance(v, list):
                filter_conds.append(model_field.in_(v))
            else:
                if model_field.expression.type.__visit_name__ == 'string':
                    filter_conds.append(model_field.contains(str(v)))
                else:
                    filter_conds.append(model_field==v)
    return filter_conds, ', '.join(err_messages)

def dump_model_data(model: ModelBase, columns: list = None):
    values = {}
    if not columns:
        columns, _ = model_columns(model)
    hidden_fields = []
    if hasattr(model, '__hidden_response_fields__') and isinstance(getattr(model, '__hidden_response_fields__', None), list):
        hidden_fields = getattr(model, '__hidden_response_fields__')
    for c in columns:
        if c in hidden_fields:
            continue
        if hasattr(model, c):
            val = getattr(model, c)
            if isinstance(val, bytes):
                val = val.decode('utf-8')
            values[c] = val
    return values

def read_request_parameters(request: tornado.httputil.HTTPServerRequest):
    params = args_as_dict(request)
    if request.method == 'POST':
        if request.headers['Content-Type'] == 'application/json':
            filters2 = request_body_as_json(request)
            for k, v in filters2.items():
                # if body contains the same key besides the query arguments, overwrite it.
                # it means that the post body parameter has the high priority value
                params[k] = v
        else:
            # post body parameters would already parsed in request.arguments
            pass
    if not params:
        params = {}
    return params

def get_locale_params(request: tornado.httputil.HTTPServerRequest):
    locale_params = {}
    language = ''
    if request.headers.get("Accept-Language"):
        languages = request.headers.get("Accept-Language").split(",")
        locales = []
        for language in languages:
            parts = language.strip().split(";")
            if len(parts) > 1 and parts[1].strip().startswith("q="):
                try:
                    score = float(parts[1].strip()[2:])
                    if score < 0:
                        raise ValueError()
                except (ValueError, TypeError):
                    score = 0.0
            else:
                score = 1.0
            if score > 0:
                locales.append((parts[0], score))
        if locales:
            locales.sort(key=lambda pair: pair[1], reverse=True)
            codes = [loc[0] for loc in locales]
            language = codes[0]
    else:
        # for k, v in request.cookies.items():
        #     print('cookie keys:', k, v)
        language = request.cookies.get('locale', None)
        if not language:
            language = request.headers.get('Locale', None)
    if language:
        locale_params['locale'] = str(language)
    return locale_params

def get_listquery_pager_info(params, default_list_limit=20, max_list_limit=1000):
    limit = default_list_limit
    page = 0
    if 'pagesize' in params or 'pageSize' in params:
        limit = toint(params['pagesize' if 'pagesize' in params else 'pageSize'])
        if not limit:
            limit = default_list_limit
        elif limit > max_list_limit:
            limit = max_list_limit
    if 'page' in params or 'current' in params:
        page = toint(params['page' if 'page' in params else 'current']) - 1
        if page < 0:
            page = 0
    offset = page * limit
    return limit, offset

def get_listquery_sort_info(params):
    order = ''
    direction = 'asc'
    sortsParam = params
    if 'sorts' in params and isinstance(params['sorts'], dict):
        sortsParam = params['sorts']
    elif 'sorts' in params and isinstance(params['sorts'], str):
        try:
            sortsParam = json.loads(params['sorts'])
        except:
            pass
    elif 'sorter' in params and isinstance(params['sorter'], str):
        try:
            sortsParam = json.loads(params['sorter'])
            for k, v in sortsParam.items():
                order = k
                direction = 'asc' if v.startswith('asc') else 'desc'
                return order, direction
        except:
            pass

    if 'sort' in sortsParam and isinstance(sortsParam['sort'], str):
        order = sortsParam['sort']
    direction = 'asc'
    for sortOrderKey in ['direction', 'order']:
        if sortOrderKey in sortsParam and isinstance(sortsParam[sortOrderKey], str):
            direction = sortsParam[sortOrderKey].lower()
            if direction != 'desc':
                direction = 'asc'
            break
    if order and order.startswith('-'):
        order = order[1:]
        direction = 'desc'
    return order, direction



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import traceback
import logging
from hawthorn.dbproxy import DbProxy
from hawthorn.modelutils import meta_data
from .model.supports import Migration, MigrationProgress

APP_NAME = 'Unknown'
MIGRATES = []
LOG = logging.getLogger('migrations.main')

if os.getenv('APP_NAME'):
    APP_NAME = os.getenv('APP_NAME', 'Unknown')

def set_appname(appname=''):
    global APP_NAME
    if appname and isinstance(appname, str):
        APP_NAME = appname

def set_migrates(migrates=[]):
    global MIGRATES
    if migrates and isinstance(migrates, list):
        MIGRATES = migrates

async def create_tables():
    # ensure table were created
    dbinstance = DbProxy().get_model_dbinstance(Migration)
    async with dbinstance.engine.begin() as conn:
        await conn.run_sync(meta_data.create_all)
        LOG.info('ensure all table meta data were created')
        # meta_data.create_all(conn.sync_connection)

async def do_migrations():
    await create_tables()
    for fn in MIGRATES:
        version = migration_parse_version(fn.__name__)
        if not version:
            continue
        cur_migration_executed = await migration_check_version(version)
        if cur_migration_executed:
            continue
        migration_result = await fn()
        if migration_result is not False:
            await migration_finish_version(version)
            LOG.info('migrates by version:%s finished.' % (version))

async def migration_check_version(version):
    obj = await DbProxy().find_item(Migration, {Migration.version==version})
    if obj and obj.status > 0:
        return True
    return False

async def migration_finish_version(version):
    obj = await DbProxy().find_item(Migration, {Migration.app==APP_NAME, Migration.version==version})
    if not obj:
        obj = Migration()
        obj.set_session_uid(APP_NAME)
        obj.app = APP_NAME
        obj.version = version
    else:
        obj.set_session_uid(APP_NAME)
        obj.obsoleted = 0
    obj.status = 1
    if obj.id:
        await DbProxy().update_item(obj, auto_flush=True)
    else:
        await DbProxy().insert_item(obj, auto_flush=True)
    return True

def migration_parse_version(fnName):
    slices = fnName.split('_')
    if len(slices) < 1:
        return ''
    return slices[1]

async def migration_get_progress(module, total_cb=None):
    tb = traceback.extract_stack()
    version = migration_parse_version(tb[-2][2])

    p = await DbProxy().find_item(MigrationProgress, {MigrationProgress.app==APP_NAME, MigrationProgress.version==version, MigrationProgress.module==module})
    if not p:
        p = MigrationProgress()
        p.set_session_uid(APP_NAME)
        p.app = APP_NAME
        p.version = version
        p.progress = 0
        p.module = module
        p.value = ''
        p.total = total_cb() if total_cb else 0
        p.cost_seconds = 0
        p.left_seconds = 0
        await DbProxy().insert_item(p, auto_flush=True)
    else:
        p.set_session_uid(APP_NAME)
    
    return p

async def migration_update_progress(progressRecord, curProcessed, value, dt):
    progressRecord.progress += curProcessed
    progressRecord.value = str(value)
    progressRecord.cost_seconds += int(dt)
    speed = float(progressRecord.progress / progressRecord.cost_seconds) if progressRecord.cost_seconds else 1.0
    progressRecord.left_seconds = int((progressRecord.total - progressRecord.progress) / speed)
    await DbProxy().update_item(progressRecord)
    return progressRecord



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import tornado.web
import tornado.httputil
from wtforms import Form
from hawthorn.asynchttphandler import GeneralTornadoHandler, routes
from hawthorn.modelutils import ModelBase, meta_data, get_model_class_name, model_columns
from .accesscontrol.authorizedsession import authorized_session_access_control
from .modelapihandlers import ModelRESTfulHandler
from .modelrelationhandlers import ModelRelationsRESTfulHandler
from .pagedescriptorhandlers import ModelPageDescriptorsApiHandler

LOG = logging.getLogger('lycium.restfulwrapper.register')

def find_model_class_by_table_name(table_name: str):
    tbl_model: ModelBase = None
    for model in ModelBase.registry.mappers:
        if hasattr(model.class_, '__tablename__') and model.class_.__tablename__ == table_name:
            # print(model.class_.__name__)
            if model.class_.__name__[0] != '_':
                tbl_model = model.class_
                break
    return tbl_model

def find_model_class_by_cls_name(cls_name: str):
    for model in ModelBase.registry.mappers:
        if model.class_.__name__ == cls_name:
            return model.class_
    return None

def register_model_general_api_handlers(model: ModelBase, endpoint: str = '', form: Form = None, web_app: tornado.web.Application=None, **options):
    # handler = ModelApiHandler(model=model, add_form=add_form, edit_form=edit_form)
    ac = options.pop('ac', [])
    if not ac:
        ac = [authorized_session_access_control]
    if not endpoint:
        endpoint = '/' + str(get_model_class_name(model)).lower() + 's'
    if endpoint.endswith('/'):
        endpoint.rstrip('/')
    if not endpoint.startswith('/'):
        endpoint = '/' + endpoint

    uri = endpoint+r'(?:/(?P<id>\w+))?'
    local_routes = [
        (uri, ModelRESTfulHandler, dict(model=model, form=form, ac=ac))
    ]
    LOG.info('registing RESTful endpoint %s', uri)
    if hasattr(model, '_sa_class_manager'):
        for attr in model._sa_class_manager.local_attrs.values():
            if attr.prop.strategy_wildcard_key == 'relationship' and attr.prop.secondary:
                tbl = meta_data.tables.get(attr.prop.secondary, None)
                if tbl is not None:
                    middle_model = find_model_class_by_table_name(attr.prop.secondary)
                    src_field = ''
                    dst_field = ''
                    src_model = model
                    dst_model = find_model_class_by_cls_name(attr.prop.argument) if isinstance(attr.prop.argument, str) else attr.prop.argument
                    if not middle_model:
                        continue
                    columns, pk = model_columns(middle_model)
                    for colname in columns:
                        if colname == pk:
                            continue
                        col = getattr(middle_model, colname)
                        if col.expression.foreign_keys:
                            for k in col.expression.foreign_keys:
                                tbl_cols = k.target_fullname.split('.')
                                if len(tbl_cols) < 2:
                                    continue
                                if hasattr(model, '__tablename__') and tbl_cols[0] == model.__tablename__:
                                    src_field = col.key
                                else:
                                    dst_field = col.key
                                break
                    if middle_model is not None and src_field and dst_field and dst_model is not None:
                        uri = endpoint + r'/(?P<instanceID>\w+)/' + attr.key
                        local_routes.append((uri, ModelRelationsRESTfulHandler, dict(middle_model=middle_model, src_field=src_field, dst_field=dst_field, src_model=src_model, dst_model=dst_model)))
                        LOG.info('registing relations RESTful endpoint %s', uri)
    # local_routes.append((uri_prefix+'/get', GeneralTornadoHandler, dict(callback=handler.handler_get, methods=['GET'], ac=ac)))
    # local_routes.append((uri_prefix+'/list', GeneralTornadoHandler, dict(callback=handler.handler_list, methods=['GET'], ac=ac)))
    # local_routes.append((uri_prefix+'/add', GeneralTornadoHandler, dict(callback=handler.handler_add, methods=['POST'], ac=ac)))
    # local_routes.append((uri_prefix+'/edit', GeneralTornadoHandler, dict(callback=handler.handler_edit, methods=['PATCH'], ac=ac)))
    # local_routes.append((uri_prefix+'/delete', GeneralTornadoHandler, dict(callback=handler.handler_delete, methods=['DELETE'], ac=ac)))

    if web_app:
        [web_app.add_handlers(route[0], route) for route in local_routes]
    else:
        routes.routes.extend(local_routes)

# def register_middle_relation_model_api_handlers(model: ModelBase, uri_prefix: str, src_field: str, dst_field: str, src_model: ModelBase, dst_model: ModelBase, web_app: tornado.web.Application=None, **options):
#     handler = ModelRelationsApiHandler(model, src_field, dst_field, src_model, dst_model)
#     ac = options.pop('ac', [])
#     if not ac:
#         ac = [authorized_session_access_control]
#     if not uri_prefix:
#         uri_prefix = '/' + str(get_model_class_name(model)).lower() + 's'
#     if uri_prefix.endswith('/'):
#         uri_prefix.rstrip('/')
#     if not uri_prefix.startswith('/'):
#         uri_prefix = '/' + uri_prefix

#     local_routes = []
#     local_routes.append((uri_prefix+'/list', GeneralTornadoHandler, dict(callback=handler.handler_relation_ids, methods=['GET'], ac=ac)))
#     local_routes.append((uri_prefix+'/add', GeneralTornadoHandler, dict(callback=handler.handler_add, methods=['POST'], ac=ac)))
#     local_routes.append((uri_prefix+'/update', GeneralTornadoHandler, dict(callback=handler.handler_update, methods=['PATCH'], ac=ac)))
#     local_routes.append((uri_prefix+'/delete', GeneralTornadoHandler, dict(callback=handler.handler_delete, methods=['DELETE'], ac=ac)))

#     if web_app:
#         [web_app.add_handlers(route[0], route) for route in local_routes]
#     else:
#         routes.routes.extend(local_routes)

def register_model_page_descriptor_api_handler(get_page_descriptor_model, web_app: tornado.web.Application=None, **options):
    handler = ModelPageDescriptorsApiHandler(get_page_descriptor_model)
    ac = options.pop('ac', [])
    if not ac:
        ac = [authorized_session_access_control]
    local_routes = [
        ('/api/pages/descriptors', GeneralTornadoHandler, dict(callback=handler.handler_get, methods=['GET'], ac=ac))
    ]
    if web_app:
        [web_app.add_handlers(route[0], route) for route in local_routes]
    else:
        routes.routes.extend(local_routes)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from hawthorn.utilities import Constant

class _ResultCode(Constant):
    # 通过
    OK = 0
    # 失败
    FAIL = -1
    # 参数非法
    INVALID_PARAM = -2
    # 数据不存在
    DATA_DOES_NOT_EXISTS = 103
    # 数据已存在
    DAT_ALREADY_EXISTS = 104
    # 没有新数据可保存
    NOTHING_TO_SAVE = 105
    INTERNAL_ERROR = 500

    # 密码错误
    PASSWORD_NOT_CORRECT = 110011
    # 账户过期
    ACCOUNT_EXPIRED = 110012
    # 账户被冻结
    ACCOUNT_WERE_FROZEN = 110013

RESULT_CODE = _ResultCode()



#!/usr/bin/python
#-*- coding:utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lycium-rest",
    version="0.0.1",
    author="kevinyjn",
    author_email="kevinyjn@gmail.com",
    description="common python programing encapsulation library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/libpub/lycium-rest",
    packages=setuptools.find_packages(exclude=[".tests", ".tests.", "tests.*", "tests"]),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "pycrypto",
        "sqlalchemy==1.4.25",
        "IPy",
        "requests",
        "tornado",
        "hawthorn",
        "PyJWT",
        "python-i18n[YAML]"
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import tornado.web
import tornado.httputil
import sqlalchemy
import i18n
from http import HTTPStatus
from typing import Iterable
from hawthorn.asynchttphandler import GeneralTornadoHandler, request_body_as_json
from hawthorn.session import TornadoSession
from hawthorn.modelutils import ModelBase, model_columns, get_model_class_name
from hawthorn.dbproxy import DbProxy
from ..valueobjects.resultcodes import RESULT_CODE
from ..valueobjects.responseobject import GeneralResponseObject, ListResponseObject

from .utils import format_model_query_conditions, read_request_parameters, get_locale_params

LOG = logging.getLogger('services.generalmodelapi.apihandlers')

class ModelRelationsRESTfulHandler(tornado.web.RequestHandler):
    """
    Model API many-to-many relationships handler wrapper
    """
    def initialize(self, middle_model: ModelBase, src_field: str, dst_field: str, src_model: ModelBase, dst_model: ModelBase, ac=[]):
        self.middle_model = middle_model
        self.src_field = src_field
        self.dst_field = dst_field
        self.src_model = src_model
        self.dst_model = dst_model
        if not getattr(self.middle_model, self.src_field):
            LOG.error('initialize %s middle relations api handler failed, there is no %s field in model', get_model_class_name(self.middle_model), self.src_field)
        if not getattr(self.middle_model, self.dst_field):
            LOG.error('initialize %s middle relations api handler failed, there is no %s field in model', get_model_class_name(self.middle_model), self.dst_field)
        self._ac = []
        self.session = TornadoSession(self)
        self.set_access_control(ac)

    def set_access_control(self, ac):
        self._ac = []
        if ac:
            if isinstance(ac, list):
                for f in ac:
                    if callable(f):
                        self._ac.append(['', f])
            elif isinstance(ac, dict):
                for k, f in ac:
                    if callable(f):
                        self._ac.append([k, f])
            elif callable(ac):
                self._ac.append(['', ac])

    def pre_check_middle_model_fields(self, **kwargs):
        if not hasattr(self.middle_model, self.src_field) or not hasattr(self.middle_model, self.dst_field):
            err_message = '%s or %s filed does not exists in relation model' % (self.src_field, self.dst_field)
            return err_message
        return ''

    async def get(self, instanceID: str|int = None):
        """
        API handler of get many-to-many relationships ids
        """
        if not instanceID:
            self.set_status(HTTPStatus.NOT_FOUND)
            self.finish()
            return
        locale_params = get_locale_params(self.request)
        err_message = self.pre_check_middle_model_fields(**locale_params)
        if err_message:
            LOG.error('failed to execute handler of get %s relations id, %s', get_model_class_name(self.middle_model), err_message)
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR, err_message)
            self.finish()
            return
        filters = read_request_parameters(self.request)
        filters[self.src_field] = instanceID
        dst_field: sqlalchemy.Column = None
        if self.src_field in filters:
            dst_field = getattr(self.middle_model, self.dst_field)
        elif self.dst_field in filters:
            dst_field = getattr(self.middle_model, self.src_field)
        
        result = ListResponseObject(RESULT_CODE.DATA_DOES_NOT_EXISTS, message=i18n.t('basic.data_not_exists', **locale_params))
        while result.code != RESULT_CODE.OK:
            filter_conds, err_msg = format_model_query_conditions(self.middle_model, filters=filters)
            if err_msg:
                result.code = RESULT_CODE.INVALID_PARAM
                result.message = err_msg
                break

            limit = 1000
            offset = 0
            rows, total = await DbProxy().query_list(self.middle_model, filters=filter_conds, limit=limit, offset=offset, selections=[dst_field.name])
            result.code = RESULT_CODE.OK
            result.message = i18n.t('basic.success', **locale_params)
            result.total = total
            result.data = rows
            if limit:
                result.pageSize = limit
                result.page = int(offset/limit) + 1
        self.set_header('Content-Type', 'application/json')
        self.write(result.encode_json())
        self.finish()

    async def post(self, instanceID: str|int = None):
        """
        API handler of add many-to-many relationships
        """
        await self.common_update_relations(instanceID, 'add')

    async def patch(self, instanceID: str|int = None):
        """
        API handler of update many-to-many relationships ids
        """
        await self.common_update_relations(instanceID, 'update')

    async def delete(self, instanceID: str|int = None):
        """
        API handler of delete delete many-to-many relationships ids
        """
        if not instanceID:
            self.set_status(HTTPStatus.NOT_FOUND)
            self.finish()
            return
        locale_params = get_locale_params(self.request)
        err_message = self.pre_check_middle_model_fields(**locale_params)
        if err_message:
            LOG.error('failed to execute handler of delete %s relations id, %s', get_model_class_name(self.middle_model), err_message)
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR, err_message)
            self.finish()
            return
        inputs = request_body_as_json(self.request)
        inputs[self.src_field] = instanceID
        src_value = inputs.get(self.src_field, 0)
        dst_value = inputs.get(self.dst_field, 0)
        result = GeneralResponseObject(RESULT_CODE.INVALID_PARAM, i18n.t('basic.invalid_param', **locale_params))
        while result.code != RESULT_CODE.OK:
            check_fields = [(self.src_model, self.src_field, src_value), (self.dst_model, self.dst_field, dst_value)]
            check_failed = False
            for cks in check_fields:
                if not cks[2]:
                    result.code = RESULT_CODE.INVALID_PARAM
                    result.message = '%s field should have a valid value' % (cks[1])
                    LOG.warning('delete relation ids for %s failed, %s', get_model_class_name(self.middle_model), result.message)
                    check_failed = True
                    break
            if check_failed:
                break

            try:
                del_conds = [
                    self.format_relation_find_conditions(self.middle_model, self.src_field, src_value)[0],
                    self.format_relation_find_conditions(self.middle_model, self.dst_field, dst_value)[0]
                ]
                res = await DbProxy().del_items(self.middle_model, del_conds)
                if res:
                    result.code = RESULT_CODE.OK
                    result.message = i18n.t('basic.success', **locale_params)
                    LOG.info('delete relation ids for %s with %s:%s and %s:%s succeed, affected %d rows', get_model_class_name(self.middle_model), self.src_field, str(src_value), self.dst_field, str(dst_value), res)
                else:
                    result.code = RESULT_CODE.DATA_DOES_NOT_EXISTS
                    result.message = i18n.t('basic.data_not_exists', **locale_params)
                    LOG.warning('delete relation ids for %s with %s:%s and %s:%s affected zero rows', get_model_class_name(self.middle_model), self.src_field, str(src_value), self.dst_field, str(dst_value))
            except Exception as e:
                LOG.error('delete relation ids for %s with %s:%s and %s:%s failed with error:%s', get_model_class_name(self.middle_model), self.src_field, str(src_value), self.dst_field, str(dst_value), str(e))
                result.code = RESULT_CODE.FAIL
                result.message = str(e)
            break
        self.set_header('Content-Type', 'application/json')
        self.write(result.encode_json())
        self.finish()

    async def common_update_relations(self, instanceID: str|int, mode: str):
        """
        common relation ids operator of many-to-many relationships ids
        """
        if not instanceID:
            self.set_status(HTTPStatus.NOT_FOUND)
            self.finish()
            return
        locale_params = get_locale_params(self.request)
        err_message = self.pre_check_middle_model_fields(**locale_params)
        if err_message:
            LOG.error('failed to execute handler of %s %s relations id, %s', mode, get_model_class_name(self.middle_model), err_message)
            return err_message, HTTPStatus.INTERNAL_SERVER_ERROR
        inputs = request_body_as_json(self.request)
        inputs[self.src_field] = instanceID
        src_value = inputs.get(self.src_field, 0)
        dst_value = inputs.get(self.dst_field, 0)
        extra_field_values = {}
        middle_model_columns, middle_model_pk = model_columns(self.middle_model)
        for k in middle_model_columns:
            if k == middle_model_pk:
                continue
            if k in inputs and k != self.src_field and k != self.dst_field:
                extra_field_values[k] = inputs[k]

        result = GeneralResponseObject(RESULT_CODE.INVALID_PARAM, i18n.t('basic.invalid_param', **locale_params))
        while result.code != RESULT_CODE.OK:
            check_models = [(self.src_model, self.src_field, src_value), (self.dst_model, self.dst_field, dst_value)]
            check_failed = False
            for cks in check_models:
                if not cks[2]:
                    result.code = RESULT_CODE.INVALID_PARAM
                    result.message = '%s field should have a valid value' % (cks[1])
                    LOG.warning('%s relation ids for %s failed, %s', mode, get_model_class_name(self.middle_model), result.message)
                    check_failed = True
                    break
                if cks[0]:
                    try:
                        _, pk = model_columns(cks[0])
                        check_conds = self.format_relation_find_conditions(cks[0], pk, cks[2])
                        src_exists = await DbProxy().get_count(cks[0], check_conds)
                        if not src_exists:
                            result.code = RESULT_CODE.DATA_DOES_NOT_EXISTS
                            result.message = get_model_class_name(cks[0]) + f' by {cks[2]} does not exists'
                            LOG.warning('add relation ids for %s failed, %s', get_model_class_name(self.middle_model), result.message)
                            check_failed = True
                            break
                    except Exception as e:
                        result.code = RESULT_CODE.INVALID_PARAM
                        result.message = str(e)
                        LOG.warning('add relation ids for %s failed, %s', get_model_class_name(self.middle_model), result.message)
                        check_failed = True
                        break
            if check_failed:
                break

            adding_datas = []
            exists_records = {}
            src_ids = [v for v in src_value] if isinstance(src_value, Iterable) else [src_value]
            dst_ids = [v for v in dst_value] if isinstance(dst_value, Iterable) else [dst_value]
            for v1 in src_ids:
                for v2 in dst_ids:
                    adding_datas.append((v1, v2))
            if 'add' != mode:
                try:
                    exists_rows = await DbProxy().query_all(self.middle_model, self.format_relation_find_conditions(self.middle_model, self.src_field, src_value))
                    for ele in exists_rows:
                        exists_records[f'{ele[self.src_field]}-{ele[self.dst_field]}'] = ele
                except Exception as e:
                    result.code = RESULT_CODE.INVALID_PARAM
                    result.message = str(e)
                    LOG.warning('%s relation ids for %s while find existing records by %s:%s failed, %s', mode, get_model_class_name(self.middle_model), self.src_field, str(src_value), result.message)
                    break
            add_records = []
            for ele in adding_datas:
                m = self.middle_model()
                if hasattr(m, 'set_session_uid'):
                    session_uid = self.session['user_id']
                    setattr(m, 'set_session_uid', session_uid)
                setattr(m, self.src_field, ele[0])
                setattr(m, self.dst_field, ele[1])
                for k, v in extra_field_values.items():
                    setattr(m, k, v)
                add_records.append(m)
                existing_key = f'{ele[0]}-{ele[1]}'
                if existing_key in exists_records:
                    del exists_records[existing_key]

            del_records_conds = []
            del_ids = [row[middle_model_pk] for row in exists_records.values()]
            if del_ids:
                del_records_conds.append(getattr(self.middle_model, middle_model_pk).in_(del_ids))

            try:
                LOG.info('%s relation ids for %s adding records with %s:%s and %s:%s', mode, get_model_class_name(self.middle_model), self.src_field, str(src_value), self.dst_field, str(dst_value))
                res = await DbProxy().insert_items(add_records, auto_flush=True)
                if res:
                    result.code = RESULT_CODE.OK
                    result.message = i18n.t('basic.success', **locale_params)
                    LOG.info('%s relation ids for %s adding records with %s:%s and %s:%s succeed', mode, get_model_class_name(self.middle_model), self.src_field, str(src_value), self.dst_field, str(dst_value))

                    if del_ids:
                        LOG.info('%s relation ids for %s deleting old relation records %s', mode, get_model_class_name(self.middle_model), str(del_ids))
                        del_count = await DbProxy().del_items(self.middle_model, del_records_conds)
                        LOG.info('%s relation ids for %s deleting old relation records %s affected %d rows', mode, get_model_class_name(self.middle_model), str(del_ids), del_count)
                else:
                    LOG.error('%s relation ids for %s with %s:%s and %s:%s failed', mode, get_model_class_name(self.middle_model), self.src_field, str(src_value), self.dst_field, str(dst_value))
            except Exception as e:
                LOG.error('%s relation ids for %s with %s:%s and %s:%s failed with error:%s', mode, get_model_class_name(self.middle_model), self.src_field, str(src_value), self.dst_field, str(dst_value), str(e))
                result.code = RESULT_CODE.FAIL
                result.message = str(e)
            break
        self.set_header('Content-Type', 'application/json')
        self.write(result.encode_json())
        self.finish()

    def format_relation_find_conditions(self, model: ModelBase, field_name:str, find_value):
        conds = []
        if isinstance(find_value, dict) or isinstance(find_value, list) or isinstance(find_value, tuple):
            conds.append(getattr(model, field_name).in_(find_value))
        else:
            conds.append(getattr(model, field_name)==find_value)
        return conds



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sqlalchemy import Column, Integer, BigInteger, SmallInteger, String

from hawthorn.modelutils import ModelBase
from hawthorn.modelutils.behaviors import ModifyingBehevior

class _MigrationTable(ModelBase):
    __tablename__ = 'sys_migration'

    id = Column('id', Integer, primary_key=True, autoincrement=True)
    app = Column('app', String(64), index=True)
    version = Column('version', String(64), index=True)
    status = Column('status', SmallInteger, index=True)

class Migration(_MigrationTable, ModifyingBehevior):
    """
    Migration information
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
